from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, TypeVar

from typing_extensions import Protocol

if TYPE_CHECKING:
    from mypy_boto3_dynamodb import DynamoDBClient, DynamoDBServiceResource
    from mypy_boto3_s3 import S3Client, S3ServiceResource
    from mypy_boto3_secretsmanager import SecretsManagerClient
    from mypy_boto3_sqs import SQSClient

T = TypeVar("T", covariant=True)


class FromEnv(Protocol[T]):
    def from_env(self) -> T:
        pass


class FromEnvException(Exception):
    pass


ClientGetParams = NamedTuple(
    "ClientGetParams",
    (
        ("boto3_client_name", str),  # e.g. "s3" or "sqs"
        ("endpoint_url_key", str),  # e.g. "SQS_ENDPOINT"
        ("access_key_id_key", str),
        ("access_key_secret_key", str),
    ),
)


def _client_get(client_create_fn: Callable[..., Any], params: ClientGetParams) -> Any:
    """
    :param client_create_fn: the `boto3.client` or `boto3.resource` function
    """
    which_service = params.boto3_client_name
    endpoint_url = os.getenv(params.endpoint_url_key)
    access_key_id = os.getenv(params.access_key_id_key)
    access_key_secret = os.getenv(params.access_key_secret_key)

    # AWS_REGION is Fargate-specific, most AWS stuff uses AWS_DEFAULT_REGION.
    region = os.getenv("AWS_REGION") or os.getenv("AWS_DEFAULT_REGION")
    if not region:
        raise FromEnvException("Please set AWS_REGION= or AWS_DEFAULT_REGION=")

    # Not needed long term, more to help migrate to `env_helpers`.
    # Notably, when `is_local` is not set, it won't break anything.
    is_local = os.getenv("IS_LOCAL", None)

    # Unlike Rust FromEnv, we rely on boto3's built in region handling.

    if all((endpoint_url, access_key_id, access_key_secret)):
        # Local, all are passed in from docker-compose.yml
        logging.info(f"Creating a local client for {which_service}")
        assert (
            is_local != False
        ), f"You must pass in credentials for a local {which_service} client"
        return client_create_fn(
            params.boto3_client_name,
            endpoint_url=endpoint_url,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=access_key_secret,
            region_name=region,
        )
    elif endpoint_url and not any((access_key_id, access_key_secret)):
        # Local or AWS doing cross-region stuff
        return client_create_fn(
            params.boto3_client_name,
            endpoint_url=endpoint_url,
            region_name=region,
        )
    elif not any((endpoint_url, access_key_id, access_key_secret)):
        # AWS
        logging.info("Creating a prod client")
        assert (
            is_local != True
        ), f"You can't pass in credentials for a prod {which_service} client"
        return client_create_fn(
            params.boto3_client_name,
            region_name=region,
        )
    else:
        raise FromEnvException(
            f"You specified access key but not endpoint for {params.boto3_client_name}?"
        )


_SQSParams = ClientGetParams(
    boto3_client_name="sqs",
    endpoint_url_key="SQS_ENDPOINT",
    access_key_id_key="SQS_ACCESS_KEY_ID",
    access_key_secret_key="SQS_ACCESS_KEY_SECRET",
)


class SQSClientFactory(FromEnv["SQSClient"]):
    def __init__(self, boto3_module: Any):
        self.client_create_fn = boto3_module.client

    def from_env(self) -> SQSClient:
        client: SQSClient = _client_get(self.client_create_fn, _SQSParams)
        return client


_S3Params = ClientGetParams(
    boto3_client_name="s3",
    endpoint_url_key="S3_ENDPOINT",
    access_key_id_key="S3_ACCESS_KEY_ID",
    access_key_secret_key="S3_ACCESS_KEY_SECRET",
)


class S3ClientFactory(FromEnv["S3Client"]):
    def __init__(self, boto3_module: Any):
        self.client_create_fn = boto3_module.client

    def from_env(self) -> S3Client:
        client: S3Client = _client_get(self.client_create_fn, _S3Params)
        return client


class S3ResourceFactory(FromEnv["S3ServiceResource"]):
    def __init__(self, boto3_module: Any):
        self.client_create_fn = boto3_module.resource

    def from_env(self) -> S3ServiceResource:
        client: S3ServiceResource = _client_get(self.client_create_fn, _S3Params)
        return client


_DynamoDBParams = ClientGetParams(
    boto3_client_name="dynamodb",
    endpoint_url_key="DYNAMODB_ENDPOINT",
    access_key_id_key="DYNAMODB_ACCESS_KEY_ID",
    access_key_secret_key="DYNAMODB_ACCESS_KEY_SECRET",
)


class DynamoDBResourceFactory(FromEnv["DynamoDBServiceResource"]):
    def __init__(self, boto3_module: Any):
        self.client_create_fn = boto3_module.resource

    def from_env(self) -> DynamoDBServiceResource:
        client: DynamoDBServiceResource = _client_get(
            self.client_create_fn, _DynamoDBParams
        )
        return client


class DynamoDBClientFactory(FromEnv["DynamoDBClient"]):
    def __init__(self, boto3_module: Any):
        self.client_create_fn = boto3_module.client

    def from_env(self) -> DynamoDBClient:
        client: DynamoDBClient = _client_get(self.client_create_fn, _DynamoDBParams)
        return client


_SecretsManagerParams = ClientGetParams(
    boto3_client_name="secretsmanager",
    endpoint_url_key="SECRETSMANAGER_ENDPOINT",
    access_key_id_key="SECRETSMANAGER_ACCESS_KEY_ID",
    access_key_secret_key="SECRETSMANAGER_ACCESS_KEY_SECRET",
)


class SecretsManagerClientFactory(FromEnv["SecretsManagerClient"]):
    def __init__(self, boto3_module: Any):
        self.client_create_fn = boto3_module.client

    def from_env(self) -> SecretsManagerClient:
        client: SecretsManagerClient = _client_get(
            self.client_create_fn, _SecretsManagerParams
        )
        return client



import inspect
import logging
import os
import sys


def get_module_grapl_logger(default_log_level: str = "ERROR") -> logging.Logger:
    """
    Callers should put
    LOGGER = get_module_grapl_logger()
    at module scope.
    """
    caller_stack = inspect.stack()[1]
    caller_module = inspect.getmodule(caller_stack[0])
    assert caller_module
    logger = logging.getLogger(caller_module.__name__)
    logger.setLevel(os.getenv("GRAPL_LOG_LEVEL", default_log_level))
    # While a lot of our code defines this, I believe it just doubles our log output
    # logger.addHandler(logging.StreamHandler(stream=sys.stdout))
    return logger



"""
This is reasonably well-documented in `vscode_debugger.rst`.

I arbitrarily chose `debugpy`, the VS Code debugger (formerly known as ptsvd).
It'd be pretty easy to add the PyCharm/IntelliJ debugger too (which uses pydevd)
"""
import logging
import os
import subprocess
import sys


def _install_from_pip(package: str) -> None:
    # Gross, but the suggested way to install from pip mid-python-program!
    # Doing it this way so that <every executable that depends on grapl_common> doesn't inherently
    # need to import debugpy
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", "--upgrade", package]
    )


# TODO: we should probably have some fixed 'services' StrEnum somewhere
SERVICE_TO_PORT = {
    # As you need to debug more services, add more services here.
    # Make sure you expose the port in the appropriate docker-compose file.
    "grapl_e2e_tests": 8400,
    "analyzer_executor": 8401,
}


def _should_debug_service(service: str) -> bool:
    """
    When you set
    DEBUG_SERVICES=grapl_e2e_tests
    you'll start a debug listener on the `grapl_e2e_tests`

    When you set
    DEBUG_SERVICES=grapl_e2e_tests,some_future_service
    you'll start two debug listeners - 1 for each service
    """
    env_var = os.getenv("DEBUG_SERVICES")
    if not env_var:
        return False
    debug_services = set(env_var.split(","))
    return service in debug_services


def wait_for_vsc_debugger(service: str) -> None:
    if not _should_debug_service(service):
        return
    port = SERVICE_TO_PORT.get(service, None)
    if not port:
        logging.error("Couldn't find a debug port for service {service}.")
        return

    _install_from_pip("debugpy")
    import debugpy  # type: ignore

    assert (
        8400 <= port < 8500
    ), "84xx range is reserved for our debuggers. You likely want 1 per service."
    host = "0.0.0.0"
    logging.info(f">> Debugpy listening for client at {host}:{port}")
    debugpy.listen((host, port))
    debugpy.wait_for_client()
    logging.info(f">> Debugpy connected!")



import re
from typing import Pattern, Sequence, Union

from typing_extensions import Final, Literal

DEFAULT_SAMPLE_RATE: Final[float] = 1.0

_INVALID_CHARS: Final[Pattern] = re.compile(r"[|#,=:]")

MetricType = Literal["g", "c", "ms", "h"]  # |m is also valid, but I chose to ignore it


class TagPair:
    tag_key: str
    tag_value: str

    def __init__(self, tag_key: str, tag_value: str) -> None:
        _reject_invalid_chars(tag_key)
        _reject_invalid_chars(tag_value)
        self.tag_key = tag_key
        self.tag_value = tag_value

    def statsd_serialized(self) -> str:
        return ":".join((self.tag_key, self.tag_value))


def _reject_invalid_chars(s: str) -> None:
    # TODO - consider a cache of acceptable strings, since python inters its strings
    match = _INVALID_CHARS.search(s)
    if match:
        raise ValueError(f"Invalid character in input {s}")


def statsd_format(
    metric_name: str,
    value: Union[int, float],
    typ: MetricType,
    sample_rate: float = DEFAULT_SAMPLE_RATE,
    tags: Sequence[TagPair] = (),
) -> str:
    """
    Mainline `statsd` hasn't chosen a tag syntax yet: https://github.com/statsd/statsd/issues/619
    However, it looks like they will be supporting the Graphite and DogStatsD formats.
    I've arbitrarily chosen the DogStatsD format.

    You can find the spec here:
    https://github.com/b/statsd_spec

    <METRIC_NAME>:<VALUE>|<TYPE>|@<SAMPLE_RATE>|#<TAG_KEY_1>:<TAG_VALUE_1>,<TAG_2>
    """
    _reject_invalid_chars(metric_name)

    # sections will eventually be joined by |
    sections = [f"{metric_name}:{value}", typ]

    # Add sample rate.
    # Counter - 'c' - is the only metric that responds to sample rate
    if typ == "c" and sample_rate != DEFAULT_SAMPLE_RATE:
        if not (0.0 <= sample_rate <= 1.0):
            raise ValueError(f"Bad sample rate {sample_rate}")
        sections.append(f"@{sample_rate}")

    # Add tags
    tag_section = ",".join(tag_pair.statsd_serialized() for tag_pair in tags)
    if tag_section:
        sections.append(f"#{tag_section}")
    return "|".join(sections)



from datetime import datetime, timedelta
from typing import NewType

# we have a bunch of bare ints around the codebase
# representing timestamps, this makes it slightly easier
# to reason about
Millis = NewType("Millis", int)

# and for millisecond-durations that have nothing to do with Unix Time:
MillisDuration = NewType("MillisDuration", int)


def as_datetime(millis: Millis) -> datetime:
    return datetime.fromtimestamp(millis / 1000.0)


def as_millis(dt: datetime) -> Millis:
    return Millis(int(dt.timestamp() * 1000))


def as_millis_duration(delta: timedelta) -> MillisDuration:
    return MillisDuration(int(delta.microseconds / 1000))




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/grapl-common

from setuptools import setup

setup(**{
    'author': 'Grapl, Inc.',
    'author_email': 'grapl-code@graplsecurity.com',
    'classifiers': [
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    'description': 'Shared code between Grapl components',
    'license': 'MIT',
    'long_description': """# grapl-common
Shared library between different Grapl python services.

Ideally things that would go in here would be things like:
* serializable and deserializable types (since those cross service boundaries)
* common utilities, i.e. dealing with datetime stuff
""",
    'maintainer': 'Grapl, Inc.',
    'maintainer_email': 'grapl-code@graplsecurity.com',
    'name': 'grapl-common',
    'project_urls': {
        'Documentation': 'https://grapl.readthedocs.io',
        'Source': 'https://github.com/grapl-security/grapl',
        'Tracker': 'https://github.com/grapl-security/grapl/issues',
    },
    'url': 'https://github.com/grapl-security/grapl',
    'version': '0.0.5',
    'package_dir': {
        '': 'src',
    },
    'packages': (
        'grapl_common',
        'grapl_common.debugger',
        'grapl_common.metrics',
    ),
    'namespace_packages': (
    ),
    'package_data': {
        'grapl_common': (
            'py.typed',
        ),
    },
    'install_requires': (
        'mypy_boto3_dynamodb',
        'mypy_boto3_s3',
        'mypy_boto3_secretsmanager',
        'mypy_boto3_sqs',
        'typing_extensions',
    ),
})



from contextlib import contextmanager
from datetime import datetime, timezone
from enum import Enum
from sys import stdout
from typing import Callable, Iterator, Optional, Sequence, TextIO, Union

from grapl_common.metrics.statsd_formatter import (
    DEFAULT_SAMPLE_RATE,
    MetricType,
    TagPair,
    statsd_format,
)
from grapl_common.time_utils import MillisDuration, as_millis_duration
from typing_extensions import Protocol


class Writeable(Protocol):
    """
    Protocol for `stdout`.
    Like a TextIO but with less to mock out.
    """

    def write(self, some_str: str) -> int:
        pass


_RESERVED_UNIT_TAG = "_unit"


class HistogramUnit(str, Enum):
    MILLIS = "millis"
    MICROS = "micros"
    SECONDS = "seconds"


class MetricReporter:
    """
    Print metrics to stdout that are picked up by Metric Forwarder later.
    Prefer MetricReporter.create(my_service_name) to get an instance.
    """

    def __init__(
        self,
        service_name: str,
        utc_now: Callable[[], datetime],
        out: Writeable,
    ):
        self.service_name = service_name
        self.utc_now = utc_now
        self.out = out

    @staticmethod
    def create(service_name: str) -> "MetricReporter":
        return MetricReporter(service_name, utc_now=MetricReporter._utcnow, out=stdout)

    def write_metric(
        self,
        metric_name: str,
        value: Union[int, float],
        typ: MetricType,
        sample_rate: float = DEFAULT_SAMPLE_RATE,
        tags: Sequence[TagPair] = (),
    ) -> None:
        now = self.utc_now()
        now_ts = self._format_time_for_cloudwatch(now)
        statsd = statsd_format(
            metric_name,
            value,
            typ,
            sample_rate,
            tags,
        )
        self.out.write(f"MONITORING|{self.service_name}|{now_ts}|{statsd}\n")

    def counter(
        self,
        metric_name: str,
        value: Union[int, float],
        sample_rate: float = DEFAULT_SAMPLE_RATE,
        tags: Sequence[TagPair] = (),
    ) -> None:
        """
        Sort of like a gauge, but with deltas.
        Metrics sent will increment or decrement the value of the gauge rather than giving its current value.
        """
        self.write_metric(
            metric_name=metric_name,
            value=value,
            sample_rate=sample_rate,
            typ="c",
            tags=tags,
        )

    def gauge(
        self,
        metric_name: str,
        value: Union[int, float],
        tags: Sequence[TagPair] = (),
    ) -> None:
        """
        An instantaneous measurement of a value, like the gas gauge in a car.
        """
        self.write_metric(
            metric_name=metric_name,
            value=value,
            typ="g",
            tags=tags,
        )

    def histogram(
        self,
        metric_name: str,
        value: MillisDuration,
        tags: Sequence[TagPair] = (),
        unit: Optional[HistogramUnit] = None,
    ) -> None:
        """
        A histogram is a measure of the distribution of timer values over time, calculated at the
        server. As the data exported for timers and histograms is the same,
        this is currently an alias for a timer.

        example: the time to complete rendering of a web page for a user.
        """
        if unit:
            tags = [TagPair(_RESERVED_UNIT_TAG, unit.value)] + [t for t in tags]
        self.write_metric(
            metric_name=metric_name,
            value=value,
            typ="h",
            tags=tags,
        )

    @contextmanager
    def histogram_ctx(
        self,
        metric_name: str,
        tags: Sequence[TagPair] = (),
    ) -> Iterator[None]:
        """
        A histogram is a measure of the distribution of timer values over time, calculated at the
        server. As the data exported for timers and histograms is the same,
        this is currently an alias for a timer.

        example: the time to complete rendering of a web page for a user.
        """
        start = self.utc_now()
        try:
            yield
        finally:
            end = self.utc_now()
            value = as_millis_duration(end - start)
            self.write_metric(
                metric_name=metric_name,
                value=value,
                typ="h",
                tags=tags,
            )

    _TIME_SPEC = "milliseconds"

    @staticmethod
    def _format_time_for_cloudwatch(dt: datetime) -> str:
        return dt.isoformat(timespec=MetricReporter._TIME_SPEC)

    @staticmethod
    def _utcnow() -> datetime:
        return datetime.now(timezone.utc)
